/***************************************************************************
 *   Copyright (C) 2008 by David Sansome                                   *
 *   me@davidsansome.com                                                   *
 *                                                                         *
 *   Copyright (C) 2010 by Cédric Bellegarde                               *
 *   gnumdk@gmail.com                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <QImage>
#include <QDesktopWidget>
#include <QFile>
#include "engine.h"
#include "dummyapplication.h"
#include "utilities.h"
#include "qcolorutils.h"
#include "rcproperties.h"

#include <time.h>

#define IS_APP(app) (strstr(m_wmClass, app))

#define ENGINE_SETUP(OptionClassName, imageFill, isButton, Palette) \
    QImage *image = new QImage(m_size, QImage::Format_ARGB32_Premultiplied); \
    QPalette palette;\
    if(isButton) palette = *m_palette_b;\
    else palette = *m_palette;\
    if(imageFill)\
    {\
        QPalette::ColorGroup colorGroup = (m_state == GTK_STATE_INSENSITIVE) ? QPalette::Inactive : QPalette::Normal; \
        image->fill(palette.color(colorGroup, Palette).rgb()); \
    }\
    else\
    {\
        image->fill(qRgba(0, 0, 0, 0));\
    }\
    QPainter p(image);\
    OptionClassName option; \
    setupOption(&option, palette);

#define ENGINE_SETUP_IMAGE(OptionClassName, isButton, image)\
    QPalette palette;\
    if(isButton) palette = *m_palette_b;\
    else palette = *m_palette;\
    QPainter p(image);\
    OptionClassName option; \
    setupOption(&option, palette);

#define ENGINE_END_PAINTING \
    p.end();

#define ENGINE_DRAW\
    cairo_surface_t *surface = cairo_image_surface_create_for_data(image->bits(), CAIRO_FORMAT_ARGB32 , m_size.width(), m_size.height(),image->bytesPerLine());\
    cairo_t *t = gdk_cairo_create(m_window);\
    cairo_set_source_surface(t, surface, m_topLeft.x(), m_topLeft.y());\
    cairo_paint(t);\
    cairo_surface_destroy(surface);\
    cairo_destroy(t);


#define ENGINE_DELETE_OBJECTS\
    if(image)\
        delete image;\
     
#define ENGINE_FINISH\
    ENGINE_END_PAINTING\
    ENGINE_DRAW\
    ENGINE_DELETE_OBJECTS\
     
void destroyEngine()
{
    // Firefox fix
    delete QApplication::instance();
}

static int dummyErrorHandler(Display*, XErrorEvent*)
{
    return 0;
}

static int dummyIOErrorHandler(Display*)
{
    return 0;
}



Engine* Engine::s_instance = NULL;

Engine::Engine()
    : m_enabled(true),
      m_debug(false),
      m_window(NULL),
      m_style(NULL)
{
    // If this environment variable is set then lots of debugging info will be printed to the console
    m_debug = (getenv("GTK_OXYGEN_ENGINE_DEBUG") != NULL) ? 1 : 0;

    s_instance = this; // Don't put GTK_QT_DEBUG* calls before this line

    GTK_QT_DEBUG_FUNC


    m_wmClass = gdk_get_program_class();
    m_noGradients = IS_APP("Firefox") || IS_APP("Thunderbird") || IS_APP("Seamonkey");

    // Disable for bad apps
    if(! strcmp(m_wmClass, "Soffice") ||
       ! strcmp(m_wmClass, "<unknown>")) {
        m_enabled = false;
        qDebug() << "Disabling Oxygen engine for" << m_wmClass;
    }

    // Also disable the engine if the user has requested it
    if(getenv("GTK_OXYGEN_ENGINE_DISABLE") != NULL) {
        m_enabled = false;
        qDebug() << "Disabling Oxygen engine as requested by GTK_OXYGEN_ENGINE_DISABLE";
    }

    if(!m_enabled) {
        QString gtkrc;
        QFile file;
        if(file.exists("/usr/share/themes/QtCurve/gtk-2.0/gtkrc")) {
            gtkrc = "/usr/share/themes/QtCurve/gtk-2.0/gtkrc";
        } else {
            gtkrc = "/usr/share/themes/Clearlooks/gtk-2.0/gtkrc";
        }
        gchar *default_files[] = {gtkrc.toAscii().data(), NULL};
        gtk_rc_set_default_files(default_files);
        return;
    }

    // Stop Qt from integrating with the Glib event loop and stealing events from GTK
    setenv("QT_NO_GLIB", "1", 1);
    // Stop Qt session management
    char* sessionManager = getenv("SESSION_MANAGER");
    setenv("SESSION_MANAGER", "", 1);

    // Create a QApplication if there isn't one already
    if(QApplication::instance() == NULL) {
        // The QApplication constructor sets its own X11 error handlers.
        // We want to save the current (GTK) ones so we can restore them after
        int (*originalErrorHandler)(Display*, XErrorEvent*) = XSetErrorHandler(dummyErrorHandler);
        int (*originalIOErrorHandler)(Display*) = XSetIOErrorHandler(dummyIOErrorHandler);

        display = gdk_display_get_default();
        g_signal_connect(G_OBJECT(display), "closed", G_CALLBACK(destroyEngine), NULL);

        // Construct our Qt application - needed to use QPainter
        // The dummy application has empty implementations for Qt's session management functions
        // We don't need them, and they try to use Qt's event loop (which doesn't work)
        new DummyApplication(gdk_x11_display_get_xdisplay(display));

        // Restore GTK's error handlers
        XSetErrorHandler(originalErrorHandler);
        XSetIOErrorHandler(originalIOErrorHandler);
    }
    
    // Restore session manager env var
    setenv("SESSION_MANAGER", sessionManager, 1);
    
    m_qtStyle = QApplication::setStyle("oxygen");

    if(!m_qtStyle) {
        m_qtStyle = QApplication::setStyle("plastique");
        qDebug() << "Oxygen style not installed or i'm a static binary";
    }

    m_palette = NULL;
    m_palette_b = NULL;

    // Create some Qt widgets here so we don't have to recreate them in every draw* function
    m_dummyWidget = new QWidget(0);
    m_dummyButton = new QPushButton(m_dummyWidget);
    m_dummyCheckBox = new QCheckBox(m_dummyWidget);
    m_dummyRadioButton = new QRadioButton(m_dummyWidget);
    m_dummyTabBar = new QTabBar(m_dummyWidget);
    m_dummyTabWidget = new QTabWidget(m_dummyWidget);
    m_dummyLineEdit = new QLineEdit(m_dummyWidget);
    m_dummyMenu = new QMenu(m_dummyWidget);
    m_dummyMenuBar = new QMenuBar(m_dummyWidget);
    m_dummyComboBox = new QComboBox(m_dummyWidget);
    m_dummySlider = new QSlider(m_dummyWidget);
    m_dummyScrollBar = new QScrollBar(m_dummyWidget);
    m_dummyToolBar = new QToolBar(m_dummyWidget);

    // Initialize various GTK RC properties
    RcProperties::setRcProperties();
    
    m_cache.setMaxItems(5);
}

Engine::~Engine()
{
    GTK_QT_DEBUG_FUNC

    if(!m_enabled)
        return;

    delete m_dummyWidget; // This will also clean up all the other dummy widgets which are its children

    delete QApplication::instance();

    delete m_palette;
    delete m_palette_b;

    s_instance = NULL;
}

void Engine::setupOption(QStyleOption* option, const QPalette& palette) const
{
    option->direction = m_dummyWidget->layoutDirection();
    option->rect = QRect(QPoint(0, 0), m_size);
    option->palette = palette;
    option->fontMetrics = m_dummyWidget->fontMetrics();
    // Setup state
    option->state = QStyle::State_Enabled | QStyle::State_Active | QStyle::State_Raised;

    switch(m_state) {
    case GTK_STATE_ACTIVE:
        option->state &= ~QStyle::State_Raised;
        option->state |= QStyle::State_Sunken;
        break;
    case GTK_STATE_PRELIGHT:
        option->state |= QStyle::State_MouseOver;
        break;
    case GTK_STATE_SELECTED:
        option->state |= QStyle::State_HasFocus;
        break;
    case GTK_STATE_INSENSITIVE:
        option->state &= ~QStyle::State_Enabled;
        break;
    case GTK_STATE_NORMAL:
    default:
        break;
    }

    if(m_hasFocus)
        option->state |= QStyle::State_HasFocus;
}

void Engine::drawButton(bool defaultButton)
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionButton, false, true, QPalette::Window)
    // Handle default buttons (ones that get activated when you press enter)
    option.features = defaultButton ? QStyleOptionButton::DefaultButton : QStyleOptionButton::None;

    // change palette button background color to deal with button position in main window
    option.palette.setColor(QPalette::Button, QColorUtils::backgroundColor(option.palette.color(QPalette::Button), m_topLeft.y(), m_toplevelSize.height()));

    // draw
    m_qtStyle->drawControl(QStyle::CE_PushButton, &option, &p, m_dummyButton);

    ENGINE_FINISH
}

void Engine::drawToolButton()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionToolButton, false, true, QPalette::Window)
    option.state |= QStyle::State_AutoRaise;
    option.subControls &= ~QStyle::SC_ToolButtonMenu;
    m_qtStyle->drawComplexControl(QStyle::CC_ToolButton, &option, &p, m_dummyToolBar);

    ENGINE_FINISH
}

void Engine::drawCheckBox(bool checked)
{
    GTK_QT_DEBUG_FUNC

    // Hack for checkbox in list (bad)
    if(m_size.width() < 23) {
        m_size = QSize(23, 23);
        m_topLeft -= QPoint(4, 4);
    }

    // Handle checked state
    m_dummyCheckBox->setChecked(checked);

    ENGINE_SETUP(QStyleOptionButton, false, true, QPalette::Window)
    option.state |= checked ? QStyle::State_On : QStyle::State_Off;
    // Without this, clicking a checkbox looks weird as you lose the mouseover highlight for a second
    if(m_state == GTK_STATE_ACTIVE)
        option.state |= QStyle::State_MouseOver;
    m_qtStyle->drawControl(QStyle::CE_CheckBox, &option, &p, m_dummyCheckBox);
    ENGINE_FINISH
}

// This code sux, don't understand gtk behaviour
void Engine::drawRadioButton(bool checked)
{
    GTK_QT_DEBUG_FUNC

    // Hack for radio in list (bad)
    if(m_size.width() < 23) {
        m_size = QSize(23, 23);
        m_topLeft -= QPoint(4, 4);
    }

    // Handle checked state
    m_dummyRadioButton->setChecked(checked);

    ENGINE_SETUP(QStyleOptionButton, false, true, QPalette::Window)
    option.state |= checked ? QStyle::State_On : QStyle::State_Off;
    // Without this, clicking a radiobutton looks weird as you lose the mouseover highlight for a second
    if(m_state == GTK_STATE_ACTIVE)
        option.state |= QStyle::State_MouseOver;

    m_qtStyle->drawControl(QStyle::CE_RadioButton, &option, &p, m_dummyRadioButton);
    ENGINE_FINISH
}

void Engine::drawTab(QTabWidget::TabPosition position)
{
    GTK_QT_DEBUG_FUNC

    switch(position) {
    case QTabWidget::West:
        m_size += QSize(9, 0);
        break;
    case QTabWidget::East:
        m_size += QSize(9, 0);
        break;
    default:
        m_size += QSize(0, 9);
    }

    ENGINE_SETUP(QStyleOptionTabV3, false, false, QPalette::Window)
    option.cornerWidgets = QStyleOptionTab::NoCornerWidgets;
    option.position = QStyleOptionTab::Middle;
    option.selectedPosition = QStyleOptionTab::NotAdjacent;

    switch(position) {
    case QTabWidget::West:
        option.shape = QTabBar::RoundedWest;
        break;
    case QTabWidget::East:
        option.shape = QTabBar::RoundedEast;
        break;
    case QTabWidget::South:
        option.shape = QTabBar::RoundedSouth;
        break;
    default:
        option.shape = QTabBar::RoundedNorth;
    }

    // All GTK tabs are drawn in the ACTIVE state except the active tab which is drawn in the NORMAL state...
    if(m_state == GTK_STATE_NORMAL)
        option.state |= QStyle::State_Selected;

    m_qtStyle->drawControl(QStyle::CE_TabBarTab, &option, &p, m_dummyTabBar);

    ENGINE_END_PAINTING

    QRect rect;
    QImage *copy;
    switch(position) {
    case QTabWidget::West:
        if(m_state == GTK_STATE_NORMAL)
            m_topLeft += QPoint(-1, -1);
        else
            m_topLeft += QPoint(-2, -1);
        m_size -= QSize(5, 0);
        break;
    case QTabWidget::East:
        if(m_state == GTK_STATE_NORMAL)
            m_topLeft += QPoint(-3, -1);
        else
            m_topLeft += QPoint(-1, -1);
        rect = QRect(QPoint(5, 0), m_size);
        copy = new QImage(image->copy(rect));
        delete image;
        image = copy;
        break;
    case QTabWidget::South:
        if(m_state == GTK_STATE_NORMAL)
            m_topLeft -= QPoint(1, 4);
        else
            m_topLeft -= QPoint(1, 0);

        rect = QRect(QPoint(0, 5), m_size);
        copy = new QImage(image->copy(rect));
        delete image;
        image = copy;
        break;
    default:
        m_topLeft -= QPoint(1, 0);
        if(m_state == GTK_STATE_NORMAL)
            m_size -= QSize(0, 13);
        else
            m_size -= QSize(0, 15);
    }
    ENGINE_DRAW
    ENGINE_DELETE_OBJECTS
}

/*
void Engine::drawTabFrame(unsigned long xid)
{
    GTK_QT_DEBUG_FUNC
    QImage *image = background(QRect(m_topLeft, m_size), xid);
    if(image) {
        ENGINE_SETUP_IMAGE(QStyleOptionTabWidgetFrame, false, image)
        option.shape = QTabBar::RoundedNorth;
        option.leftCornerWidgetSize = QSize(0, 0);
        option.rightCornerWidgetSize = QSize(0, 0);
        option.tabBarSize = QSize(0, 0);

        m_qtStyle->drawPrimitive(QStyle::PE_FrameTabWidget, &option, &p, m_dummyTabWidget);
        ENGINE_FINISH
    }
}
*/

void Engine::drawLineEdit()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOption, true, false, QPalette::Base)
    ENGINE_FINISH
}

void Engine::drawLineEditFrame(int tx, int ty, bool isMozilla)
{
    QImage *image = NULL;
    GTK_QT_DEBUG_FUNC
    if(isMozilla) {
        ENGINE_SETUP(QStyleOptionFrameV2, false, false, QPalette::Window)
        option.lineWidth = m_qtStyle->pixelMetric(QStyle::PM_DefaultFrameWidth, &option, m_dummyLineEdit);
        option.midLineWidth = 0;
        option.state |= QStyle::State_Sunken;
        option.state &= ~QStyle::State_Raised;
        m_qtStyle->drawPrimitive(QStyle::PE_FrameLineEdit, &option, &p, m_dummyLineEdit);
        ENGINE_END_PAINTING

        // Another horrible hack for spinbuttons :(
        QPainter i(image);
        i.setPen(Qt::transparent);
        QPalette::ColorGroup colorGroup = QPalette::Normal;
        i.setBrush(QBrush(palette.color(colorGroup, QPalette::Base).rgb(), Qt::SolidPattern));
        i.drawRect(QRectF(QPoint(option.lineWidth + 1, option.lineWidth + 1), QSize(m_size.width() - (option.lineWidth + 1) * 2, m_size.height() - (option.lineWidth + 1) * 2)));
        i.end();
        ENGINE_DRAW
        ENGINE_DELETE_OBJECTS
    } else {
        image = background(QPoint(tx, ty));
        ENGINE_SETUP_IMAGE(QStyleOptionFrameV2, false, image)
        option.lineWidth = m_qtStyle->pixelMetric(QStyle::PM_DefaultFrameWidth, &option, m_dummyLineEdit);
        option.midLineWidth = 0;
        option.state |= QStyle::State_Sunken;
        option.state &= ~QStyle::State_Raised;
        m_qtStyle->drawPrimitive(QStyle::PE_FrameLineEdit, &option, &p, m_dummyLineEdit);
        ENGINE_END_PAINTING

        // Another horrible hack for spinbuttons :(
        QPainter i(image);
        i.setPen(Qt::transparent);
        QPalette::ColorGroup colorGroup = QPalette::Normal;
        i.setBrush(QBrush(palette.color(colorGroup, QPalette::Base).rgb(), Qt::SolidPattern));
        i.drawRect(QRectF(QPoint(option.lineWidth + 1, option.lineWidth + 1), QSize(m_size.width() - (option.lineWidth + 1) * 2, m_size.height() - (option.lineWidth + 1) * 2)));
        i.end();
        ENGINE_DRAW
        ENGINE_DELETE_OBJECTS
    }
}

void Engine::drawFrame(int type, int tx, int ty, QTabWidget::TabPosition position)
{
    GTK_QT_DEBUG_FUNC
    QSize origSize = m_size;

    if(position == QTabWidget::West || position == QTabWidget::East) {
        drawRectFrame(tx, ty);
    } else {
        QImage *image = background(QPoint(tx, ty));

        ENGINE_SETUP_IMAGE(QStyleOptionFrameV2, false, image)
        option.state |= QStyle::State_Sunken;
        option.state &= ~QStyle::State_Raised;

        QStyle::PrimitiveElement element;
        switch(type) {
        case 0:  element = QStyle::PE_FrameGroupBox; break;
        case 1:
        default: element = QStyle::PE_Frame;         break;
        }

        switch(position) {
        case QTabWidget::South:
            p.translate(m_size.width(), m_size.height());
            p.rotate(180);
            break;
        default:
            m_topLeft -= QPoint(0, 2);
            break;
        }

        m_qtStyle->drawPrimitive(element, &option, &p, m_dummyWidget);

        ENGINE_FINISH
    }
}

void Engine::drawArrow(GtkArrowType direction)
{
    GTK_QT_DEBUG_FUNC

    QStyle::PrimitiveElement element;
    switch(direction) {
    case GTK_ARROW_UP:    element = QStyle::PE_IndicatorArrowUp;    break;
    case GTK_ARROW_LEFT:  element = QStyle::PE_IndicatorArrowLeft;  break;
    case GTK_ARROW_RIGHT: element = QStyle::PE_IndicatorArrowRight; break;
    case GTK_ARROW_DOWN:
    default:              element = QStyle::PE_IndicatorArrowDown;  break;
    }

    ENGINE_SETUP(QStyleOption, false, false, QPalette::Window)

    m_qtStyle->drawPrimitive(element, &option, &p);
    ENGINE_FINISH
}

void Engine::drawListHeader()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionHeader, true, false, QPalette::Window)
    option.orientation = Qt::Horizontal;
    option.section = 1;
    option.position = QStyleOptionHeader::Middle;
    option.selectedPosition = QStyleOptionHeader::NotAdjacent;
    option.sortIndicator = QStyleOptionHeader::None;
    option.state &= ~QStyle::State_Raised;
    option.state |= QStyle::State_Sunken;

    m_qtStyle->drawControl(QStyle::CE_Header, &option, &p, m_dummyButton);

    ENGINE_FINISH
}

void Engine::drawMenuBarItem()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionMenuItem, false, false, QPalette::Window)
    option.state |= QStyle::State_Sunken | QStyle::State_Selected;
    option.state &= ~QStyle::State_Raised;

    m_qtStyle->drawControl(QStyle::CE_MenuBarItem, &option, &p);

    ENGINE_FINISH
}

void Engine::drawMenuFrame()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionFrameV3, false, false, QPalette::Window)
    option.state = QStyle::State_Raised;
    option.lineWidth = 3;
    option.midLineWidth = 3;
    option.frameShape = QFrame::StyledPanel;

    m_qtStyle->drawControl(QStyle::CE_ShapedFrame, &option, &p, m_dummyMenu);

    ENGINE_FINISH
}

void Engine::drawToolMenuItem(int type, bool toolbar)
{
    GTK_QT_DEBUG_FUNC

    if(toolbar)
        m_size.transpose();

    ENGINE_SETUP(QStyleOptionMenuItem, false, false, QPalette::Window)
    option.state |= QStyle::State_Sunken | QStyle::State_Selected;
    option.state &= ~QStyle::State_Raised;
    option.menuItemType = QStyleOptionMenuItem::MenuItemType(type);

    m_qtStyle->drawControl(QStyle::CE_MenuItem, &option, &p);

    ENGINE_END_PAINTING

    if(toolbar) {
        QImage *copy = image;
        QMatrix matrix;
        matrix.rotate(90);
        image = new QImage(copy->transformed(matrix));
        delete copy;
        m_size.transpose();
    }

    ENGINE_DRAW
    ENGINE_DELETE_OBJECTS
}

void Engine::drawMenuCheck()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOption, false, false, QPalette::Window)

    m_qtStyle->drawPrimitive(QStyle::PE_IndicatorMenuCheckMark, &option, &p, m_dummyMenu);

    ENGINE_FINISH
}

void Engine::drawComboBox()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionComboBox, false, true, QPalette::Window)
    option.subControls &= ~QStyle::SC_ComboBoxArrow;
    m_qtStyle->drawComplexControl(QStyle::CC_ComboBox, &option, &p, m_dummyComboBox);
    ENGINE_FINISH
}

void Engine::drawRectFrame(int tx, int ty)
{
    GTK_QT_DEBUG_FUNC
    if(m_noGradients) {
        ENGINE_SETUP(QStyleOptionFrameV3, false, true, QPalette::Window)
        option.lineWidth = 1;
        option.midLineWidth = 1;
        option.frameShape = QFrame::StyledPanel;

        m_qtStyle->drawControl(QStyle::CE_ShapedFrame, &option, &p, m_dummyMenu);

        ENGINE_FINISH
    } else {
        QImage *image = background(QPoint(tx, ty));
        ENGINE_SETUP_IMAGE(QStyleOptionFrameV3, false, image)
        option.lineWidth = 1;
        option.midLineWidth = 1;
        option.frameShape = QFrame::StyledPanel;

        m_qtStyle->drawControl(QStyle::CE_ShapedFrame, &option, &p, m_dummyMenu);

        ENGINE_FINISH
    }
}

void Engine::drawProgressBar(GtkProgressBarOrientation orientation, double percentage)
{
    GTK_QT_DEBUG_FUNC

    ENGINE_SETUP(QStyleOptionProgressBarV2, true, false, QPalette::Window)
    option.maximum = 10000;
    option.minimum = 0;

    /* Stupid gtk */
    if(percentage > 0.99)
        option.progress = 10000;
    else
        option.progress = int(percentage * 10000);

    switch(orientation) {
    case GTK_PROGRESS_LEFT_TO_RIGHT:
    case GTK_PROGRESS_RIGHT_TO_LEFT:
        option.orientation = Qt::Horizontal;
        break;
    case GTK_PROGRESS_BOTTOM_TO_TOP:
    case GTK_PROGRESS_TOP_TO_BOTTOM:
        option.orientation = Qt::Vertical;
        break;
    }
    m_qtStyle->drawControl(QStyle::CE_ProgressBar, &option, &p);

    ENGINE_FINISH
}

void Engine::drawProgressChunk()
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionProgressBarV2, false, false, QPalette::Window)
    option.maximum = 10000;
    option.minimum = 0;
    option.progress = 10000;
    
    m_qtStyle->drawControl(QStyle::CE_ProgressBarContents, &option, &p);
    ENGINE_FINISH
}

void Engine::drawSpinButton(int direction)
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionSpinBox, false, false, QPalette::Window)
    option.frame = false;

    QStyle::PrimitiveElement element;
    switch(direction) {
    case 0:  element = QStyle::PE_IndicatorArrowUp;   break;
    case 1:
    default: element = QStyle::PE_IndicatorArrowDown; break;
    }

    m_qtStyle->drawPrimitive(element, &option, &p);
    ENGINE_FINISH
}

void Engine::drawSpinFrame(int tx, int ty)
{
    // Draw a double line edit frame and / it by 2
    m_size.setWidth(m_size.width() * 2);
    QImage *image = background(QPoint(tx, ty));
    ENGINE_SETUP_IMAGE(QStyleOptionFrameV2, false, image)
    option.lineWidth = m_qtStyle->pixelMetric(QStyle::PM_DefaultFrameWidth, &option, m_dummyLineEdit);
    option.midLineWidth = 0;
    option.state |= QStyle::State_Sunken;
    option.state &= ~QStyle::State_Raised;
    m_qtStyle->drawPrimitive(QStyle::PE_FrameLineEdit, &option, &p, m_dummyLineEdit);
    ENGINE_END_PAINTING

    QImage subImage = image->copy(QRect(QPoint(m_size.width() / 2, m_topLeft.y()), QSize(m_size.width() / 2, m_size.height())));
    delete image;

    QPainter pSub(&subImage);
    pSub.setPen(Qt::transparent);
    QPalette::ColorGroup colorGroup = (m_state == GTK_STATE_INSENSITIVE) ? QPalette::Inactive : QPalette::Normal;
    pSub.setBrush(QBrush(palette.color(colorGroup, QPalette::Base).rgb(), Qt::SolidPattern));
    pSub.drawRect(QRectF(QPoint(0, 4), QSize(m_size - QSize(m_size.width() / 2 + 4, 8))));
    pSub.end();

    image = new QImage(subImage);
    m_size.setWidth(m_size.width() / 2);

    ENGINE_DRAW
    ENGINE_DELETE_OBJECTS
}

void Engine::drawSlider(GtkAdjustment* adj, GtkOrientation orientation, int inverted)
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionSlider, false, false, QPalette::Window)
    option.minimum = 0;
    option.maximum = 10000;
    option.orientation = (orientation == GTK_ORIENTATION_HORIZONTAL) ? Qt::Horizontal : Qt::Vertical;

    if(!inverted)  // Normal sliders
        option.sliderValue = int((adj->value - adj->lower) / (adj->upper - adj->lower) * 10000.0);
    else // Inverted sliders... where max is at the left/top and min is at the right/bottom
        option.sliderValue = 10000 - int((adj->value - adj->lower) / (adj->upper - adj->lower) * 10000.0);

    option.sliderPosition = option.sliderValue;

    m_qtStyle->drawComplexControl(QStyle::CC_Slider, &option, &p, m_dummySlider);

    ENGINE_FINISH
}

/* http://bugreports.qt.nokia.com/browse/QTBUG-12044
 * Here trying to bypass this bug
 * We put slider at end, draw a scrollbar
 * Then we put slider at start, draw a second scrollbar
 */
void Engine::drawScrollBar(GtkOrientation orientation)
{
    Object *o = m_cache.scrollItem(m_size);
    GTK_QT_DEBUG_FUNC

    if(o && o->m_image) {
        QImage *image = o->m_image;
        ENGINE_DRAW
    } else {
        ENGINE_SETUP(QStyleOptionSlider, m_noGradients, false, QPalette::Window)

        option.minimum = 0;
        option.maximum = 10;
        option.singleStep = 0.1;
        option.pageStep = 0.1;
        option.sliderValue = 10;
        option.sliderPosition = 10;
        option.orientation = (orientation == GTK_ORIENTATION_HORIZONTAL) ? Qt::Horizontal : Qt::Vertical;
        option.subControls &= ~QStyle::SC_ScrollBarSlider;

        // Fixes a divide by 0 error in QCommonStyle if the range (upper-lower) is 0
        if(option.maximum <= option.minimum)
            option.maximum = option.minimum + 1;

        // KStyle looks at this instead of option.orientation
        if(orientation == GTK_ORIENTATION_HORIZONTAL)
            option.state |= QStyle::State_Horizontal;

        m_qtStyle->drawComplexControl(QStyle::CC_ScrollBar, &option, &p, m_dummyScrollBar);
        option.sliderValue = 0;
        option.sliderPosition = 0;
        m_qtStyle->drawComplexControl(QStyle::CC_ScrollBar, &option, &p, m_dummyScrollBar);
        
        ENGINE_END_PAINTING
        ENGINE_DRAW
        
        m_cache.insert(m_size, image, true);
    }

}

void Engine::drawScrollBarSlider(GtkOrientation orientation)
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOptionSlider, false, false, QPalette::Window)
    option.orientation = (orientation == GTK_ORIENTATION_HORIZONTAL) ? Qt::Horizontal : Qt::Vertical;
    option.subControls = QStyle::SC_ScrollBarSlider;

    if(m_state == GTK_STATE_PRELIGHT)
        option.activeSubControls = QStyle::SC_ScrollBarSlider;

    // KStyle looks at this instead of option.orientation
    if(orientation == GTK_ORIENTATION_HORIZONTAL)
        option.state |= QStyle::State_Horizontal;

    m_qtStyle->drawComplexControl(QStyle::CC_ScrollBar, &option, &p, m_dummyScrollBar);

    ENGINE_END_PAINTING

    // Get rid of the buttons so we're left with just the scrollbar slider of the size we were asked to draw
    //  image = image.copy(sliderRect);

    ENGINE_DRAW
    ENGINE_DELETE_OBJECTS
}


void Engine::drawSplitter(GtkOrientation orientation)
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOption, false, false, QPalette::Window)

    // By "horizontal", GTK means "bar goes from left to right" and Qt means "the two panes are to the left and right of the bar"
    if(orientation != GTK_ORIENTATION_HORIZONTAL)
        option.state |= QStyle::State_Horizontal;

    m_qtStyle->drawControl(QStyle::CE_Splitter, &option, &p, m_dummyWidget);

    ENGINE_FINISH
}

void Engine::drawTooltip()
{
    ENGINE_SETUP(QStyleOption, false, false, QPalette::Window)
    m_qtStyle->drawPrimitive(QStyle::PE_PanelTipLabel, &option, &p, m_dummyLineEdit);
    ENGINE_FINISH
}

void Engine::drawItemViewItem(unsigned int round)
{
    GTK_QT_DEBUG_FUNC
    
    if(round == 0) { // left rounded (start item)
            m_size += QSize(4, 0);
            ENGINE_SETUP(QStyleOptionViewItemV4, false, false, QPalette::Window)
            option.backgroundBrush = QPalette().highlight();
            m_qtStyle->drawControl(QStyle::CE_ItemViewItem, &option, &p, m_dummyWidget);
            m_size -= QSize(4, 0);
            ENGINE_FINISH
    }
    else if(round == 1) { // none rounded (middle item)
            QImage *copy;
            m_size += QSize(8, 0);
            ENGINE_SETUP(QStyleOptionViewItemV4, false, false, QPalette::Window)
            option.backgroundBrush = QPalette().highlight();
            m_qtStyle->drawControl(QStyle::CE_ItemViewItem, &option, &p, m_dummyWidget);
            copy = image;
            image = new QImage(copy->copy(QRect(QPoint(4, 0), m_size)));
            m_size -= QSize(8, 0);
            ENGINE_FINISH
            delete copy;         
    }
    else if(round == 2) { // right rounded (end item)
            QImage *copy;
            m_size += QSize(4, 0);
            ENGINE_SETUP(QStyleOptionViewItemV4, false, false, QPalette::Window)
            option.backgroundBrush = QPalette().highlight();
            m_qtStyle->drawControl(QStyle::CE_ItemViewItem, &option, &p, m_dummyWidget);
            copy = image;
            image = new QImage(copy->copy(QRect(QPoint(4, 0), m_size)));
            m_size -= QSize(4, 0);
            ENGINE_FINISH
            delete copy;         
    }
    else { // normal item
        ENGINE_SETUP(QStyleOptionViewItemV4, false, false, QPalette::Window)
        option.backgroundBrush = QPalette().highlight();
        m_qtStyle->drawControl(QStyle::CE_ItemViewItem, &option, &p, m_dummyWidget);
        ENGINE_FINISH
    }
}

void Engine::drawRect(bool fill)
{
    GTK_QT_DEBUG_FUNC
    ENGINE_SETUP(QStyleOption, fill, false, QPalette::Window)
    ENGINE_FINISH
}

QImage* Engine::oxygenBackground(QSize size)
{
    // create new image and make transparent
    QImage *image = new QImage(size, QImage::Format_ARGB32_Premultiplied);
    image->fill(Qt::transparent);
    QPainter p(image);

    // store palette, colorgroup and color locally
    const QPalette palette = *m_palette;
    const QPalette::ColorGroup colorGroup = (m_state == GTK_STATE_INSENSITIVE) ? QPalette::Inactive : QPalette::Active;
    const QColor color(palette.color(colorGroup, QPalette::Window));

    // the hard-coded metrics are copied for kdebase/workspace/libs/oxygen/oxygenhelper.cpp
    const int gradientHeight = 64;
    const int yShift = 23;
    const int splitY(qMin(300, (3 * size.height()) / 4));
    const QRect upperRect(0, -yShift, size.width(), splitY);

    // draw upper linear gradient
    QImage* vGradient(verticalGradient(color, size.width(), splitY, 0));
    p.setCompositionMode(QPainter::CompositionMode_SourceOver);
    p.drawImage(upperRect, *vGradient);
    delete vGradient;

    // draw lower flat part
    const QRect lowerRect(0, splitY - yShift, size.width(), size.height() - splitY + yShift);
    p.fillRect(lowerRect, QColorUtils::backgroundBottomColor(color));

    // draw upper radial gradient
    const int radialW(qMin(600, size.width()));
    QRect radialRect((size.width() - radialW) / 2, 0, radialW, gradientHeight);
    radialRect.translate(0, -yShift);
    QImage* rGradient(radialGradient(color, radialW, gradientHeight));
    p.drawImage(radialRect, *rGradient);
    delete rGradient;
    p.end();
    return image;
}

void Engine::drawOxygenBackground()
{
    GTK_QT_DEBUG_FUNC
    QImage *image;
    Object *o = m_cache.item(m_size);

    if(o)
        image = o->m_image;
    else
        image = NULL;

    if(!image) {
        image = oxygenBackground(m_size);
        m_cache.insert(m_size, image, false);
    }

    ENGINE_DRAW
}



void Engine::drawWidgetBackground(int tx, int ty)
{
    GTK_QT_DEBUG_FUNC
    QImage *image = background(QPoint(tx, ty));
    ENGINE_SETUP_IMAGE(QStyleOption, false, image)
    ENGINE_FINISH
}

void Engine::drawViewportBackground(int tx, int ty)
{
    QImage *image, *copy = NULL;
    QPoint point(tx, ty);
    Object *o = m_cache.item(m_size);
    if(o) {
        image = o->m_image;
    } else {
        QSize size;
        if(m_toplevelSize.width() < point.x() + m_size.width())
            size.setWidth(point.x() + m_size.width());
        else
            size.setWidth(m_toplevelSize.width());
        if(m_toplevelSize.height() < point.y() + m_size.height())
            size.setHeight(point.y() + m_size.height());
        else
            size.setHeight(m_toplevelSize.height());
        image = oxygenBackground(size);
        m_cache.insert(m_size, image, false);
    }
    copy = new QImage(image->copy(QRect(point, m_size)));
    image = copy;
    ENGINE_SETUP_IMAGE(QStyleOption, false, image)
    ENGINE_FINISH
}

QImage* Engine::background(QPoint point)
{
    GTK_QT_DEBUG_FUNC
    QImage *image = NULL, *copy = NULL;

    if(m_noGradients) {
        ENGINE_SETUP(QStyleOption, true, false, QPalette::Window)
        m_qtStyle->drawPrimitive(QStyle::PE_Widget, &option, &p, m_dummyWidget);
        return image;
    }

    Object *o = m_cache.item(m_toplevelSize);

    if(o)
        image = o->m_image;
    else
        image = NULL;
    
    if(!image) { //Not in cache, we need to build gradient cache for toplevel window
        image = oxygenBackground(m_toplevelSize);
        m_cache.insert(m_toplevelSize, image, false);
    }
    copy = new QImage(image->copy(QRect(point, m_size)));
    return copy;
}

QImage* Engine::verticalGradient(const QColor &color, int width, int height, int offset)
{
    GTK_QT_DEBUG_FUNC
    QImage *image = new QImage(width, height, QImage::Format_ARGB32_Premultiplied);
    image->fill(qRgba(0, 0, 0, 0));

    QLinearGradient gradient(0, offset, 0, height + offset);
    gradient.setColorAt(0.0, QColorUtils::backgroundTopColor(color));
    gradient.setColorAt(0.5, color);
    gradient.setColorAt(1.0, QColorUtils::backgroundBottomColor(color));

    QPainter p(image);
    p.setCompositionMode(QPainter::CompositionMode_Source);
    p.fillRect(image->rect(), gradient);
    p.end();

    return image;
}

QImage* Engine::radialGradient(const QColor &color, int width, int height)
{
    GTK_QT_DEBUG_FUNC
    QImage *image =  new QImage(width, height, QImage::Format_ARGB32_Premultiplied);
    image->fill(qRgba(0, 0, 0, 0));

    QColor radialColor = QColorUtils::backgroundRadialColor(color);
    radialColor.setAlpha(255);

    QRadialGradient gradient(64, height - 64, 64);
    gradient.setColorAt(0, radialColor);
    radialColor.setAlpha(101);
    gradient.setColorAt(0.5, radialColor);
    radialColor.setAlpha(37);
    gradient.setColorAt(0.75, radialColor);
    radialColor.setAlpha(0);
    gradient.setColorAt(1, radialColor);

    QPainter p(image);
    p.scale(width / 128.0, 1);
    p.fillRect(QRect(0, 0, 128, height), gradient);
    p.end();
    return image;
}

const char* Engine::styleName()
{
    return m_qtStyle->metaObject()->className();
}

void Engine::destroyImage(QSize size)
{
    m_cache.remove(size);
}
