/***************************************************************************
 *   Copyright (C) 2008 by David Sansome                                   *
 *   me@davidsansome.com                                                   *
 *                                                                         *
 *   Copyright (C) 2010 by Cédric Bellegarde                               *
 *   gnumdk@gmail.com                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ENGINE_H
#define ENGINE_H

// Note: All Qt includes should go before the GTK ones to avoid name clashes
#include <QWidget>
#include <QtDebug>
#include <QPainter>
#include <QApplication>
#include <QPushButton>
#include <QStyleOption>
#include <QCheckBox>
#include <QRadioButton>
#include <QTabBar>
#include <QToolBar>
#include <QTabWidget>
#include <QLineEdit>
#include <QMenu>
#include <QMenuBar>
#include <QComboBox>
#include <QSlider>
#include <QScrollBar>

#include <gtk/gtk.h>
#include <gtk/gtkstyle.h>

#include "smartcache.h"
#include "colormapper.h"

#ifdef None
#undef None
#endif

class RcProperties;

class Engine
{
    friend class RcProperties;

public:
    ~Engine();

    inline static Engine* instance();

    inline bool isEnabled() const {
        return m_enabled;
    }
    inline bool isDebug() const {
        return m_debug;
    }

    inline QStyle* style() const {
        return m_qtStyle;
    }

    inline void setWindow(GdkWindow* window, GtkStyle* style, GtkStateType state, GdkWindow *toplevel);
    inline bool setRect(int x, int y, int w, int h);
    inline void setHasFocus(bool hasFocus) {
        m_hasFocus = hasFocus;
    }
    inline void setRgbaState(bool active) {
        m_rgba = active;
    }

    inline bool noGradients() const {
        return m_noGradients;
    }
    
    inline const char* wmClass() const {
        return m_wmClass;
    }

    void drawButton(bool defaultButton);
    void drawToolButton();
    void drawCheckBox(bool checked);
    void drawRadioButton(bool checked);
    void drawTab(QTabWidget::TabPosition position);
    void drawLineEdit();
    void drawLineEditFrame(int tx, int ty, bool isMozilla);
    void drawFrame(int type, int tx, int ty, QTabWidget::TabPosition position);
    void drawArrow(GtkArrowType direction);
    void drawListHeader();
    void drawMenuBarItem();
    void drawMenuFrame();
    void drawToolMenuItem(int type, bool toolbar);
    void drawMenuCheck();
    void drawComboBox();
    void drawRectFrame(int tx, int ty);
    void drawProgressBar(GtkProgressBarOrientation orientation, double percentage);
    void drawProgressChunk();
    void drawSpinButton(int direction);
    void drawSpinFrame(int tx, int ty);
    void drawSlider(GtkAdjustment* adj, GtkOrientation orientation, int inverted);
    void drawScrollBar(GtkOrientation orientation);
    void drawScrollBarSlider(GtkOrientation orientation);
    void drawSplitter(GtkOrientation orientation);
    void drawTooltip();
    void drawItemViewItem(unsigned int round);
    void drawRect(bool fill);
    QImage* oxygenBackground(QSize size);
    void drawOxygenBackground();
    void drawWidgetBackground(int tx, int ty);
    void drawViewportBackground(int tx, int ty);
    QImage* background(QPoint point);
    QImage* verticalGradient(const QColor &color, int width, int height, int offset);
    QImage* radialGradient(const QColor &color, int width, int height);
    const char* styleName();
    void destroyImage(QSize size);

private:
    Engine();
    void setupOption(QStyleOption* option, const QPalette& palette) const;

    // Singleton
    static Engine* s_instance;

    GdkDisplay* display;

    bool m_rgba;

    QPalette *m_palette;
    QPalette *m_palette_b;

    // Global GTK-Qt Engine state
    bool m_enabled;
    bool m_debug;
    QStyle* m_qtStyle;

    // Current GTK state
    GdkWindow* m_window;
    GtkStyle* m_style;
    GtkStateType m_state;
    QPoint m_topLeft;
    QSize m_size;
    QSize m_toplevelSize;
    bool m_hasFocus;

    // Dummy widgets
    QWidget* m_dummyWidget;
    QPushButton* m_dummyButton;
    QCheckBox* m_dummyCheckBox;
    QRadioButton* m_dummyRadioButton;
    QTabBar* m_dummyTabBar;
    QTabWidget* m_dummyTabWidget;
    QLineEdit* m_dummyLineEdit;
    QMenu* m_dummyMenu;
    QMenuBar* m_dummyMenuBar;
    QComboBox* m_dummyComboBox;
    QSlider* m_dummySlider;
    QScrollBar* m_dummyScrollBar;
    QToolBar* m_dummyToolBar;

    SmartCache m_cache;
    
    const char *m_wmClass;
    bool m_noGradients;
};

Engine* Engine::instance()
{
    if(s_instance == NULL)
        new Engine();
    return s_instance;
}

void Engine::setWindow(GdkWindow* window, GtkStyle* style, GtkStateType state, GdkWindow *toplevel)
{
    int topWidth = 0;
    int topHeight = 0;
    if(GDK_IS_DRAWABLE(toplevel))
        gdk_drawable_get_size(toplevel, &topWidth, &topHeight);
    m_window = window;
    m_style = style;
    m_state = state;
    // Some apps fail if not done here (dia for exemple)
    gdk_display_sync(display);
    m_toplevelSize = QSize(topWidth, topHeight);
    if(!m_palette) m_palette = new QPalette(ColorMapper::mapGtkToQt(style, false));
    if(!m_palette_b) m_palette_b = new QPalette(ColorMapper::mapGtkToQt(style, true));
}

bool Engine::setRect(int x, int y, int w, int h)
{
    if(w <= 1 || h <= 1)
        return false;

    m_topLeft = QPoint(x, y);
    m_size = QSize(w, h);

    return true;
}
#endif
