/* This file is part of the KDE project
 * Copyright ( C ) 2007 Matthew Woehlke <mw_triad@users.sourceforge.net>
 * Copyright ( C ) 2007 Thomas Zander <zander@kde.org>
 * Copyright ( C ) 2007 Zack Rusin <zack@kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or ( at your option ) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "qcolorutils.h"
#include <cmath>

namespace QColorUtils
{

//___________________________________________________________
// contrast
static qreal _contrast = 0.5;
static qreal _bgcontrast = 0.5;
void setContrast(qreal value)
{
    _contrast = value;
    _bgcontrast = qMin(1.0, 0.9 * _contrast / 0.7);
}

//___________________________________________________________
const qreal& contrast(void)
{
    return _contrast;
}

//___________________________________________________________
const qreal& backgroundContrast(void)
{
    return _bgcontrast;
}

//____________________________________________________________________
bool lowThreshold(const QColor &color)
{

    const QColor darker(shade(color, MidShade, 0.5));
    return luma(darker) > luma(color);

}

//____________________________________________________________________
bool highThreshold(const QColor &color)
{
    const QColor lighter(shade(color, LightShade, 0.5));
    return luma(lighter) < luma(color);
}

//_________________________________________________________________________
QColor backgroundTopColor(const QColor &color)
{
    if(lowThreshold(color)) return shade(color, MidlightShade, 0.0);
    else {
        const qreal my(luma(shade(color, LightShade, 0.0)));
        const qreal by(luma(color));
        return shade(color, (my - by) * backgroundContrast());
    }

}

//_________________________________________________________________________
QColor backgroundBottomColor(const QColor &color)
{

    const QColor midColor(shade(color, MidShade, 0.0));
    if(lowThreshold(color)) return midColor;
    else {

        const qreal by(luma(color));
        const qreal my(luma(midColor));
        return shade(color, (my - by) * backgroundContrast());

    }
}
//_________________________________________________________________________
QColor backgroundRadialColor(const QColor &color)
{

    if(lowThreshold(color)) return shade(color, LightShade, 0.0);
    else if(highThreshold(color)) return color;
    else return shade(color, LightShade, backgroundContrast());
}

//____________________________________________________________________
QColor backgroundColor(const QColor &color, int height, int y)
{

    qreal ratio(qreal(y) / qMin(300, 3 * height / 4));
    if(ratio < 0.5) {

        const qreal a(2.0 * ratio);
        return mix(backgroundTopColor(color), color, a);

    } else {

        const qreal a(2.0 * ratio - 1);
        return mix(color, backgroundBottomColor(color), a);

    }

}

//___________________________________________________________________
// luma coefficient
static const qreal yc[3] = { 0.2126, 0.7152, 0.0722 };

//___________________________________________________________________
static inline qreal mixQreal(qreal a, qreal b, qreal bias)
{
    return a + (b - a) * bias;
}

//___________________________________________________________________
static inline qreal normalize(qreal a)
{
    return (a < 1.0 ? (a > 0.0 ? a : 0.0) : 1.0);
}

//___________________________________________________________________
static inline qreal gamma(qreal n)
{
    return pow(normalize(n), 2.2);
}

//___________________________________________________________________
static inline qreal igamma(qreal n)
{
    return pow(normalize(n), 1.0 / 2.2);
}

//___________________________________________________________________
static inline qreal wrap(qreal a, qreal d = 1.0)
{
    qreal r = fmod(a, d);
    return (r < 0.0 ? d + r : (r > 0.0 ? r : 0.0));
}

//___________________________________________________________________
// hcy representation of a colors
class HCY
{

public:

    //! constructor
    HCY(const QColor& color) {

        // transparency
        a = color.alphaF();

        // luma component
        y = luma(color);

        qreal r = gamma(color.redF());
        qreal g = gamma(color.greenF());
        qreal b = gamma(color.blueF());

        // hue component
        qreal p = qMax(qMax(r, g), b);
        qreal n = qMin(qMin(r, g), b);
        qreal d = 6.0 * (p - n);
        if(n == p) h = 0.0;
        else if(r == p) h = ((g - b) / d);
        else if(g == p) h = ((b - r) / d) + (1.0 / 3.0);
        else h = ((r - g) / d) + (2.0 / 3.0);

        // chroma component
        if(r == g && g == b) c = 0.0;
        else c = qMax((y - n) / y, (p - y) / (1 - y));

    }

    //! convert back to color
    QColor qColor() const {
        // start with sane component values
        qreal _h = wrap(h);
        qreal _c = normalize(c);
        qreal _y = normalize(y);

        // calculate some needed variables
        qreal _hs = _h * 6.0, th, tm;
        if(_hs < 1.0) {

            th = _hs;
            tm = yc[0] + yc[1] * th;

        } else if(_hs < 2.0) {

            th = 2.0 - _hs;
            tm = yc[1] + yc[0] * th;

        } else if(_hs < 3.0) {

            th = _hs - 2.0;
            tm = yc[1] + yc[2] * th;

        } else if(_hs < 4.0) {

            th = 4.0 - _hs;
            tm = yc[2] + yc[1] * th;

        } else if(_hs < 5.0) {

            th = _hs - 4.0;
            tm = yc[2] + yc[0] * th;

        } else {

            th = 6.0 - _hs;
            tm = yc[0] + yc[2] * th;

        }

        // calculate RGB channels in sorted order
        qreal tn, to, tp;
        if(tm >= _y) {

            tp = _y + _y * _c * (1.0 - tm) / tm;
            to = _y + _y * _c * (th - tm) / tm;
            tn = _y - (_y * _c);

        } else {

            tp = _y + (1.0 - _y) * _c;
            to = _y + (1.0 - _y) * _c * (th - tm) / (1.0 - tm);
            tn = _y - (1.0 - _y) * _c * tm / (1.0 - tm);

        }

        // return RGB channels in appropriate order
        if(_hs < 1.0) return QColor::fromRgbF(igamma(tp), igamma(to), igamma(tn), a);
        else if(_hs < 2.0) return QColor::fromRgbF(igamma(to), igamma(tp), igamma(tn), a);
        else if(_hs < 3.0) return QColor::fromRgbF(igamma(tn), igamma(tp), igamma(to), a);
        else if(_hs < 4.0) return QColor::fromRgbF(igamma(tn), igamma(to), igamma(tp), a);
        else if(_hs < 5.0) return QColor::fromRgbF(igamma(to), igamma(tn), igamma(tp), a);
        else return QColor::fromRgbF(igamma(tp), igamma(tn), igamma(to), a);
    }

    qreal h;
    qreal c;
    qreal y;
    qreal a;

};

//___________________________________________________________________
qreal luma(const QColor &color)
{

    // RGB ratios
    return
        gamma(color.redF()) * yc[0] +
        gamma(color.greenF()) * yc[1] +
        gamma(color.blueF()) * yc[2];
}

//___________________________________________________________________
qreal contrastRatio(const QColor &c1, const QColor &c2)
{
    qreal y1 = luma(c1), y2 = luma(c2);
    if(y1 > y2) return (y1 + 0.05) / (y2 + 0.05);
    else return (y2 + 0.05) / (y1 + 0.05);
}

//___________________________________________________________________
QColor lighten(const QColor &color, qreal ky, qreal kc)
{
    HCY c(color);
    c.y = 1.0 - normalize((1.0 - c.y) * (1.0 - ky));
    c.c = 1.0 - normalize((1.0 - c.c) * kc);
    return c.qColor();
}

//___________________________________________________________________
QColor darken(const QColor &color, qreal ky, qreal kc)
{
    HCY c(color);
    c.y = normalize(c.y * (1.0 - ky));
    c.c = normalize(c.c * kc);
    return c.qColor();
}

//___________________________________________________________________
static inline QColor tintHelper(const QColor &base, const QColor &color, qreal amount)
{
    HCY result(mix(base, color, pow(amount, 0.3)));
    result.y = mixQreal(luma(base), result.y, amount);

    return result.qColor();
}

//___________________________________________________________________
QColor tint(const QColor &base, const QColor &color, qreal amount)
{
    if(amount <= 0.0) return base;
    if(amount >= 1.0) return color;
    if(isnan(amount)) return base;

    double ri = contrastRatio(base, color);
    double rg = 1.0 + ((ri + 1.0) * amount * amount * amount);
    double u = 1.0, l = 0.0;
    QColor result;
    for(int i = 12 ; i ; --i) {

        double a = 0.5 * (l + u);
        result = tintHelper(base, color, a);
        double ra = contrastRatio(base, result);
        if(ra > rg) u = a;
        else l = a;

    }

    return result;

}

//___________________________________________________________________
QColor mix(const QColor &c1, const QColor &c2, qreal bias)
{
    if(bias <= 0.0) return c1;
    if(bias >= 1.0) return c2;
    if(isnan(bias)) return c1;

    qreal r = mixQreal(c1.redF(),   c2.redF(),   bias);
    qreal g = mixQreal(c1.greenF(), c2.greenF(), bias);
    qreal b = mixQreal(c1.blueF(),  c2.blueF(),  bias);
    qreal a = mixQreal(c1.alphaF(), c2.alphaF(), bias);

    return QColor::fromRgbF(r, g, b, a);
}


//___________________________________________________________________
QColor shade(const QColor &color, ShadeRole role)
{
    return shade(color, role, _contrast);
}

//___________________________________________________________________
QColor shade(const QColor &color, ShadeRole role, qreal contrast, qreal chromaAdjust)
{

    // nan -> 1.0
    contrast = (1.0 > contrast ? (-1.0 < contrast ? contrast : -1.0) : 1.0);
    qreal y = luma(color), yi = 1.0 - y;

    if(y < 0.006) {

        // handle very dark colors (base, mid, dark, shadow == midlight, light)
        switch(role) {

        case LightShade: return shade(color, 0.05 + 0.95 * contrast, chromaAdjust);
        case MidShade: return shade(color, 0.01 + 0.20 * contrast, chromaAdjust);                case DarkShade: return shade(color, 0.02 + 0.40 * contrast, chromaAdjust);
        default: return shade(color, 0.03 + 0.60 * contrast, chromaAdjust);

        }

    } else if(y > 0.93) {

        // handle very light colors (base, midlight, light == mid, dark, shadow)
        switch(role) {

        case MidlightShade: return shade(color, -0.02 - 0.20 * contrast, chromaAdjust);
        case DarkShade: return shade(color, -0.06 - 0.60 * contrast, chromaAdjust);
        case ShadowShade: return shade(color, -0.10 - 0.90 * contrast, chromaAdjust);
        default: return shade(color, -0.04 - 0.40 * contrast, chromaAdjust);

        }

    } else {

        // handle everything else
        qreal lightAmount = (0.05 + y * 0.55) * (0.25 + contrast * 0.75);
        qreal darkAmount = (- y) * (0.55 + contrast * 0.35);
        switch(role) {
        case LightShade: return shade(color, lightAmount, chromaAdjust);
        case MidlightShade: return shade(color, (0.15 + 0.35 * yi) * lightAmount, chromaAdjust);
        case MidShade: return shade(color, (0.35 + 0.15 * y) * darkAmount, chromaAdjust);
        case DarkShade: return shade(color, darkAmount, chromaAdjust);
        default: return darken(shade(color, darkAmount, chromaAdjust), 0.5 + 0.3 * y);
        }

    }

}

//___________________________________________________________________
QColor shade(const QColor &color, qreal ky, qreal kc)
{
    HCY c(color);
    c.y = normalize(c.y + ky);
    c.c = normalize(c.c + kc);
    return c.qColor();
}


}
